#include <stdio.h>
#include <stdlib.h>
#include <cutil_inline.h>


long floatingOperationsPerKernel = 10;	// add+mul instructions in *.ptx file for clearArrays kernel

__device__ __constant__ int length;

// each thread clears array of some lenght "lenght"
extern __shared__ int k[];
__global__ void clearArrays( float* array) {
	int startDataIdx = threadIdx.x + (blockDim.x * length) * blockIdx.x;

	for(k[threadIdx.x] = 0; k[threadIdx.x]<length; k[threadIdx.x]++) {
		array[startDataIdx+blockDim.x*k[threadIdx.x]] = 0.0f;
	}

}

// device info
long memoryClock = 400000000;
long memoryInterface = 64;
long ddr = 2;
long numberOfMultiprocessors = 2;
long shaderClock = 900000000;
long multiplyAdd = 2;
long numberOfProcessors = numberOfMultiprocessors * 8;


int noBlocks = 32;
int noThreads = noBlocks * 256;
int dataPerThread = 2000;
long repeats = 100;
long arrLength = dataPerThread * noThreads;
float *hArray;
float *dArray;

void startTest() {
	cutilSafeCall(cudaMemcpyToSymbol("length", &dataPerThread, sizeof(int), 0, cudaMemcpyHostToDevice));
	for(int r=0; r<repeats; r++) {
		clearArrays<<< noBlocks, noThreads/noBlocks, noThreads/noBlocks * sizeof(int) >>>( dArray);
		cutilSafeCall( cudaThreadSynchronize() );
	}
}


void cleanArray() {
    free(hArray);
    cutilSafeCall(cudaFree(dArray));
}

void initArray() {
	printf("arrLenght: %ld bytes\n",(sizeof(float) * arrLength));
	hArray = (float*)malloc(sizeof(float) * arrLength);
	for(int i=0; i<arrLength; i++) {
		hArray[i] = (float)i;//0.0f;
	}
	dArray = NULL;
	cutilSafeCall( cudaMalloc((void**) &dArray, ( sizeof(float) * arrLength )) );
	cutilSafeCall( cudaMemcpy( dArray, hArray, ( sizeof(float) * arrLength ), cudaMemcpyHostToDevice) );

}
void printArray() {
	cutilSafeCall( cudaMemcpy( hArray, dArray, ( sizeof(float) * arrLength ), cudaMemcpyDeviceToHost) );
	for(int i=0; i<arrLength; i++) {
		printf("%f \n", hArray[i]);
	}printf("\n");
}

int main(int argc, char** argv) {

	initArray();
	unsigned int timerGPU = 0;
	CUT_SAFE_CALL( cutCreateTimer( &timerGPU));
	CUT_SAFE_CALL( cutStartTimer( timerGPU));
	startTest();
	CUT_SAFE_CALL( cutStopTimer( timerGPU));
	printf ("GPU execution time: %lf ms\n", cutGetTimerValue( timerGPU) );

	printf("CUDA error: %s\n", cudaGetErrorString(cudaGetLastError()));

	//printArray();

	float gflops = (float)(floatingOperationsPerKernel * repeats * noThreads * dataPerThread) / (float)(cutGetTimerValue( timerGPU) / 1000);
	float tGflops = (float)(numberOfProcessors * shaderClock * multiplyAdd);
	float bandwidth = (float)(repeats * arrLength * 4) / (float)(cutGetTimerValue( timerGPU) / 1000);
	float tBandwidth = (float)memoryClock * (float)memoryInterface * (float) ddr/ (float)8;

	//printf("flops calculation: %ld * %ld * %d * %d / %f\n",floatingOperationsPerKernel, repeats, noThreads, dataPerThread, (float)(cutGetTimerValue( timerGPU) / 1000));
	printf("flops: %.1f Gflop/s [%.1f%%]\n",gflops/1000000000, gflops/tGflops*100);
	printf("theoretical flops: %.1f Gflop/s\n",tGflops/1000000000);
	//printf("bandwidth calculation: %d * %ld * %d / %f \n",repeats, arrLength, 4, ((float)cutGetTimerValue( timerGPU) / 1000));
	printf("bandwidth used: %.1f GB/s [%.1f%%] \n",bandwidth/1000000000,bandwidth/tBandwidth*100);
	printf("theoretical bandwidth: %.1f GB/s\n", tBandwidth/1000000000);



	CUT_SAFE_CALL( cutDeleteTimer( timerGPU));
	cleanArray();
	cudaThreadExit();
	printf("cleaned OK\n");
}
